@@ -14,7 +14,7 @@ module Typst.Evaluate
   )
 where
 
-import Control.Monad (MonadPlus (mplus), foldM, foldM_)
+import Control.Monad (MonadPlus (mplus), foldM, foldM_, when)
 import Control.Monad.State (MonadTrans (lift))
 import Data.List (intersperse, sortOn)
 import Data.Foldable (toList)
@@ -32,7 +32,7 @@ import GHC.Generics (Generic)
 import System.FilePath (replaceFileName, takeBaseName,
                         takeFileName, takeDirectory, (</>))
 import Text.Parsec
-import Typst.Bind (destructuringBind)
+import Typst.Bind (destructuringBind, doBind)
 import Typst.Constructors (getConstructor)
 import Typst.Methods (getMethod)
 import Typst.Module.Standard (loadFileText, standardModule, symModule, getPath)
@@ -217,9 +217,6 @@ pElt = do
     Strong ms -> do
       body <- pInnerContents ms
       element "strong" Arguments {positional = [VContent body], named = OM.empty}
-    Bracketed ms -> do
-      body <- pInnerContents ms
-      pure $ (Txt "[" Seq.<| body) Seq.|> Txt "]"
     RawBlock lang txt ->
       element
         "raw"
@@ -247,6 +244,7 @@ pElt = do
                 ]
           }
     Heading level ms -> do
+      skipMany $ satisfyTok isBreak
       content <- pInnerContents ms
       element
         "heading"
@@ -480,6 +478,7 @@ evalExpr expr = applyShowRulesToVal =<<
                    val <- evalExpr y
                    case val of
                      VArray ys -> pure (xs <> ys)
+                     VNone -> pure xs
                      _ -> fail $ "Could not spread " <> show (valType val) <>
                                  " into array"
                  Reg e -> do
@@ -538,10 +537,7 @@ evalExpr expr = applyShowRulesToVal =<<
     Ident ident -> lookupIdentifier ident
     Let bind e -> do
       val <- evalExpr e
-      case bind of
-        BasicBind (Just ident) -> addIdentifier ident val
-        BasicBind Nothing -> pure ()
-        DestructuringBind parts -> destructuringBind addIdentifier parts val
+      doBind addIdentifierByExpr bind val
       pure VNone
     LetFunc name params e -> do
       val <- toFunction (Just name) params e
@@ -768,10 +764,13 @@ evalExpr expr = applyShowRulesToVal =<<
     Assign e1 e2 -> do
       val <- evalExpr e2
       case e1 of
-        Binding (BasicBind (Just ident)) -> updateIdentifier ident val
-        Binding (BasicBind Nothing) -> pure ()
-        Binding (DestructuringBind parts) ->
-          destructuringBind updateIdentifier parts val
+        Binding bind -> doBind updateExpression bind val
+        Array spreadables -> do
+          bind <- arrayExprToBind spreadables
+          doBind updateExpression bind val
+        Dict spreadables -> do
+          bind <- dictExprToBind spreadables
+          doBind updateExpression bind val
         x -> updateExpression x val
       pure VNone
     If clauses -> do
@@ -789,23 +788,16 @@ evalExpr expr = applyShowRulesToVal =<<
             case condval of
               VBoolean True -> do
                 val <- evalExpr e2
-                hadBreak <- (== FlowBreak) . evalFlowDirective <$> getState
-                joinVals result val >>= if hadBreak then pure else go
+                joinVals result val >>= ifNotBreakOrReturn go
               VBoolean False -> pure result
               _ -> fail "While loop requires a boolean condition"
-      updateState $ \st -> st {evalFlowDirective = FlowNormal}
-      go VNone
+      go VNone >>= finalizeLoop
     For bind e1 e2 -> do
       let go [] result = pure result
           go (x : xs) result = do
-            case bind of
-              BasicBind (Just ident) -> addIdentifier ident x
-              BasicBind Nothing -> pure ()
-              DestructuringBind parts ->
-                destructuringBind addIdentifier parts x
+            doBind addIdentifierByExpr bind x
             val <- evalExpr e2
-            hadBreak <- (== FlowBreak) . evalFlowDirective <$> getState
-            joinVals result val >>= if hadBreak then pure else go xs
+            joinVals result val >>= ifNotBreakOrReturn (go xs)
       source <- evalExpr e1
       items <- case source of
         VString t -> pure $ map (VString . T.singleton) (T.unpack t)
@@ -818,8 +810,7 @@ evalExpr expr = applyShowRulesToVal =<<
               )
               (OM.assocs m)
         _ -> fail $ "For expression requires an Array or Dictionary"
-      updateState $ \st -> st {evalFlowDirective = FlowNormal}
-      go items VNone
+      go items VNone >>= finalizeLoop
     Context e -> do
       evalExpr e -- TODO for now we just ignore "context"
     Return mbe -> do
@@ -895,7 +886,7 @@ toFunction mbname params e = do
               case positional as of
                 [] -> fail ("Expected parameter " <> show parts)
                 (x : xs) -> do
-                  destructuringBind addIdentifier parts x
+                  destructuringBind addIdentifierByExpr parts x
                   pure $ as {positional = xs}
             setParam as SkipParam = do
               case positional as of
@@ -1119,6 +1110,12 @@ addIdentifier ident val = do
     ((s, i) : is) -> updateState $ \st ->
       st { evalIdentifiers = (s, M.insert ident val i) : is }
 
+addIdentifierByExpr :: Monad m => Expr -> Val -> MP m ()
+addIdentifierByExpr Underscore _ = pure ()
+addIdentifierByExpr (Ident ident) val = addIdentifier ident val
+addIdentifierByExpr _ _ = fail "expected pattern, found expression"
+
+
 updateIdentifier :: Monad m => Identifier -> Val -> MP m ()
 updateIdentifier ident val = do
   let go (True, is) (s, m) = pure (True, (s, m) : is)
@@ -1149,6 +1146,23 @@ inBlock scope pa = do
       }
   pure result
 
+ifNotBreakOrReturn :: Monad m => (Val -> MP m Val) -> Val -> MP m Val
+ifNotBreakOrReturn go val = do
+  flow <- evalFlowDirective <$> getState
+  case flow of
+    FlowBreak -> pure val
+    FlowReturn _ -> pure val
+    _ -> go val
+
+finalizeLoop :: Monad m => Val -> MP m Val
+finalizeLoop result = do
+  flow <- evalFlowDirective <$> getState
+  -- do not reset FlowReturn
+  when (flow == FlowBreak || flow == FlowContinue) $
+    updateState $ \st -> st {evalFlowDirective = FlowNormal}
+  pure result
+  
+
 updateExpression :: Monad m => Expr -> Val -> MP m ()
 updateExpression
   (FuncCall (FieldAccess (Ident (Identifier "first")) e') []) val
@@ -1178,9 +1192,12 @@ updateExpression' allowNewIndices e arg val = do
     idx <- evalExpr arg
     case container of
       VArray v ->
-        case idx of
+        let toPos n = if n < 0
+              then V.length v + n
+              else n
+        in case idx of
           VInteger i ->
-            let i' = fromIntegral i
+            let i' = toPos $ fromIntegral i
             in case v V.!? i' of
                  Nothing | not allowNewIndices
                      -> fail $ "Vector does not contain index " <> show i'
@@ -1213,3 +1230,26 @@ toSelector (VRegex re) = pure $ SelectRegex re
 toSelector (VLabel t) = pure $ SelectLabel t
 toSelector (VSymbol (Symbol t _ _)) = pure $ SelectString t
 toSelector v = fail $ "could not convert " <> show v <> " to selector"
+
+arrayExprToBind :: Monad m => [Spreadable Expr] -> MP m Bind
+arrayExprToBind spredables = DestructuringBind <$> mapM spreadableToBindPart spredables
+    where
+      spreadableToBindPart (Reg expr) = Simple <$> exprToBind expr
+      spreadableToBindPart (Spr (Ident ident)) = pure $ Sink $ Just ident
+      spreadableToBindPart (Spr expr) = pure $ ExprSink expr
+
+dictExprToBind :: Monad m => [Spreadable (Expr, Expr)] -> MP m Bind
+dictExprToBind spredables = DestructuringBind <$> mapM spreadableToBindPart spredables
+    where
+      spreadableToBindPart (Reg (Ident ident, expr)) = WithKey ident <$> exprToBind expr
+      spreadableToBindPart (Reg (_, _)) = fail $ "expected identifier"
+      spreadableToBindPart (Spr (Ident ident)) = pure $ Sink $ Just ident
+      spreadableToBindPart (Spr expr) = pure $ ExprSink expr
+
+exprToBind :: Monad m => Expr -> MP m Bind
+exprToBind (Binding bind) = pure $ bind
+exprToBind Underscore = pure $ BasicBind $ Nothing
+exprToBind (Ident ident) = pure $ BasicBind $ Just ident
+exprToBind (Array spredables) = arrayExprToBind spredables
+exprToBind (Dict spredables) = dictExprToBind spredables
+exprToBind expr = pure $ ExprBind expr
