@@ -7,7 +7,7 @@ module Typst.Parse
   )
 where
 
-import Data.List (sortOn)
+import Data.List (sortOn, intercalate, dropWhileEnd)
 import Control.Applicative (some)
 import Control.Monad (MonadPlus (mzero), guard, void, when)
 import Control.Monad.Identity (Identity)
@@ -37,24 +37,24 @@ parseTypst fp inp =
     Right r -> Right r
 
 data PState = PState
-  { stIndent :: [Int],
-    stLineStartCol :: !Int,
+  { stAtStart :: !Bool,
     stAllowNewlines :: !Int, -- allow newlines if > 0
     stSpaceBefore :: Maybe (SourcePos, Text),
     stLastMathTok :: Maybe (SourcePos, Markup),
-    stContentBlockNesting :: Int
+    stContentBlockNesting :: Int,
+    stBracketNesting :: Int
   }
   deriving (Show)
 
 initialState :: PState
 initialState =
   PState
-    { stIndent = [],
-      stLineStartCol = 1,
+    { stAtStart = True,
       stAllowNewlines = 0,
       stSpaceBefore = Nothing,
       stLastMathTok = Nothing,
-      stContentBlockNesting = 0
+      stContentBlockNesting = 0,
+      stBracketNesting = 0
     }
 
 type P = Parsec Text PState
@@ -116,11 +116,17 @@ inBraces :: P a -> P a
 inBraces pa = withNewlines (between (sym "{") (char '}') pa) <* ws
 
 pMarkup :: P Markup
-pMarkup =
-  pSpace
-    <|> pHeading
-    <|> pComment
-    <|> pEol
+pMarkup = pTrivia <|> pNonTriviaMarkup
+
+pTrivia :: P Markup
+pTrivia = pTriviaWithoutEol <|> pEol
+
+pTriviaWithoutEol :: P Markup
+pTriviaWithoutEol = pSpace <|> pComment
+
+pNonTriviaMarkup :: P Markup
+pNonTriviaMarkup = do
+  mp <- pHeading
     <|> pHardbreak
     <|> pStrong
     <|> pEmph
@@ -138,20 +144,20 @@ pMarkup =
     <|> pLabelInContent
     <|> pRef
     <|> pHash
-    <|> pBracketed
     <|> pSymbol
-
--- We need to group paired brackets or the closing bracketed may be
--- taken to close a pContent block:
-pBracketed :: P Markup
-pBracketed =
-  Bracketed <$> try (between (char '[') (char ']') (many pMarkup))
+  updateState $ \st -> st {stAtStart = False}
+  pure mp
 
 pSymbol :: P Markup
 pSymbol = do
-  blockNesting <- stContentBlockNesting <$> getState
-  let isSpecial' c = isSpecial c && (c /= ']' || blockNesting == 0)
-  Text . T.singleton <$> satisfy isSpecial'
+  bracketNesting <- stBracketNesting <$> getState
+  let isSpecial' c = isSpecial c && (c /= ']' || bracketNesting > 0)
+  c <- satisfy isSpecial'
+  case c of
+    '[' -> updateState $ \st -> st { stBracketNesting = bracketNesting + 1}
+    ']' -> updateState $ \st -> st { stBracketNesting = bracketNesting - 1}
+    _ -> pure ()
+  pure $ Text $ T.singleton $ c
 
 -- equation ::= ('$' math* '$') | ('$ ' math* ' $')
 pEquation :: P Markup
@@ -433,44 +439,70 @@ mSymbol =
   lexeme ( Text . T.singleton
             <$> satisfy (\c -> not (isSpace c) && c /= '$' && c /= '\\'))
 
-withIndent :: Int -> P a -> P a
-withIndent indent pa = do
-  oldIndent <- stIndent <$> getState
-  updateState $ \st -> st {stIndent = indent : oldIndent}
-  ms <- pa
-  updateState $ \st -> st {stIndent = oldIndent}
-  pure ms
+-- Parser that discards leading trivia, and does not parse trailing trivia
+-- unless followed by non-trivia markup. Stops when non-trivia markup would
+-- match `stop` parser.
+-- Does not allow new lines except in block comments.
+pSinglelineWhileNot :: Show a => P a -> P [Markup]
+pSinglelineWhileNot stop = do
+  (concat . drop 1 . concat) <$> many pSinglelineChunk
+  where
+    pSinglelineChunk :: P [[Markup]]
+    pSinglelineChunk = try $ do
+      trivia <- many1 pTriviaWithoutEol
+      markup <- many1 (notFollowedBy stop >> pNonTriviaMarkup)
+      pure [trivia, markup]
 
 -- list ::= '-' space markup
 -- enum ::= (digit+ '.' | '+') space markup
 -- desc ::= '/' space markup ':' space markup
 pListItem :: P Markup
 pListItem = do
+  getState >>= guard . stAtStart
   col <- sourceColumn <$> getPosition
-  startLine <- stLineStartCol <$> getState
-  guard (col == startLine)
   try
     ( do
-        void $ char '-'
-        void (char ' ') <|> pBlankline
-        BulletListItem <$> withIndent col (many pMarkup)
+        char '-' *> guardAfterListOrHeadingMarker
+        BulletListItem <$> pIndented col
     )
     <|> try
       ( do
           start <- (Nothing <$ char '+') <|> (Just <$> enumListStart)
-          void (char ' ') <|> pBlankline
-          EnumListItem start <$> withIndent col (many pMarkup)
+          guardAfterListOrHeadingMarker
+          EnumListItem start <$> pIndented col
       )
-    <|> try
-      ( do
+    <|> do
           -- desc list
-          void (char '/')
-          void (many1 (char ' '))
-          term <- manyTill pMarkup (char ':')
-          skipMany spaceChar
-          optional pBlankline
-          DescListItem term <$> withIndent col (many pMarkup)
-      )
+          try $ char '/' *> guardAfterListOrHeadingMarker
+          updateState $ \st -> st {stAtStart = False}
+          term <- pSinglelineWhileNot (char ':')
+          -- previous parser might have left trivia before `:`
+          skipMany pTriviaWithoutEol <* char ':'
+          updateState $ \st -> st {stAtStart = True}
+          DescListItem term <$> pIndented col
+  where
+    -- Parser that discards leading trivia, and does not parse trailing trivia
+    -- if not followed by properly indented non-trivia markup.
+    pIndented :: Int -> P [Markup]
+    pIndented indent = do
+      (concat . drop 1 . concat) <$> many pIndentedChunk
+      where
+        pIndentedChunk :: P [[Markup]]
+        pIndentedChunk = try $ do
+          -- there might be no trivia following `:` in desc list
+          trivia <- many pTrivia
+          col <- sourceColumn <$> getPosition
+          guard $ (col > indent) || not (any isEol trivia)
+          markup <- many1 pNonTriviaMarkup
+          pure [trivia, markup]
+          where
+            isEol SoftBreak = True
+            isEol ParBreak = True
+            isEol _ = False
+
+guardAfterListOrHeadingMarker :: P ()
+guardAfterListOrHeadingMarker = lookAhead $ 
+  void (satisfy isSpace) <|> void (string "//") <|> void (string "/*") <|>  eof
 
 enumListStart :: P Int
 enumListStart = do
@@ -481,15 +513,20 @@ enumListStart = do
     Just x -> pure x
 
 -- line-comment = '//' (!unicode(Newline))*
--- block-comment = '/*' (. | block-comment)* '*/'
+-- block-comment = '/*' (block-comment | .)* ('*/' | eof)
 pComment :: P Markup
-pComment = Comment <$ (pLineComment <|> pBlockComment)
+pComment = (
+              do
+                pos <- getPosition
+                void $ string "*/"
+                setPosition pos
+                fail "Unexpected end of block comment"
+            ) <|> Comment <$ (pLineComment <|> pBlockComment)
 
 pLineComment :: P ()
 pLineComment = do
   void $ string "//"
   skipMany (satisfy (\c -> c /= '\n' && c /= '\r'))
-  void endOfLine <|> eof
 
 pBlockComment :: P ()
 pBlockComment = do
@@ -497,10 +534,9 @@ pBlockComment = do
   void $
     manyTill
       ( pBlockComment
-          <|> pLineComment
           <|> void anyChar
       )
-      (string "*/")
+      (void (string "*/") <|> eof)
 
 pSpace :: P Markup
 pSpace = Space <$ some (satisfy (\c -> isSpace c && c /= '\r' && c /= '\n'))
@@ -508,6 +544,7 @@ pSpace = Space <$ some (satisfy (\c -> isSpace c && c /= '\r' && c /= '\n'))
 pEol :: P Markup
 pEol = do
   pBaseEol
+  updateState $ \st -> st {stAtStart = True}
   (ParBreak <$ many1 pBaseEol)
     <|> (ParBreak <$ pEndOfContent)
     <|> pure SoftBreak
@@ -515,18 +552,7 @@ pEol = do
 pBaseEol :: P ()
 pBaseEol = try $ do
   void endOfLine
-  -- fail if we can't indent enough
-  indents <- stIndent <$> getState
-  case indents of
-    (i : _) -> void (try (count i (char ' '))) <|> pBlankline
-    [] -> pure ()
-  eatPrefixSpaces
-
-eatPrefixSpaces :: P ()
-eatPrefixSpaces = do
   skipMany spaceChar
-  col <- sourceColumn <$> getPosition
-  updateState $ \st -> st {stLineStartCol = col}
 
 spaceChar :: P Char
 spaceChar = satisfy (\c -> c == ' ' || c == '\t')
@@ -535,11 +561,6 @@ pHardbreak :: P Markup
 pHardbreak =
   HardBreak <$ try (char '\\' *> (void spaceChar <|> pBaseEol) *> skipMany spaceChar)
 
-pBlankline :: P ()
-pBlankline = try $ do
-  skipMany spaceChar
-  void (lookAhead endOfLine) <|> pEndOfContent
-
 pRawInline :: P Markup
 pRawInline =
   RawInline . T.pack
@@ -549,49 +570,73 @@ pRawBlock :: P Markup
 pRawBlock = do
   void $ string "```"
   numticks <- (+ 3) . length <$> many (char '`')
-  lang <- T.pack <$> (many alphaNum <* optional (char ' '))
-  optional $ try $ skipMany (char ' ') *> pEol
-  let nl = newline <* optionalGobbleIndent
-  code <-
-    T.pack
-      <$> manyTill
-        (nl <|> anyChar)
-        (string (replicate numticks '`'))
-  skipMany (char '`')
-  pure $ RawBlock lang code
+  lang <- T.pack <$> option "" pLangTag
+  code <- manyTill anyChar (string (replicate numticks '`'))
 
-optionalGobbleIndent :: P ()
-optionalGobbleIndent = do
-  indents <- stIndent <$> getState
-  case indents of
-    (i : _) -> gobble i
-    [] -> pure ()
+  -- Dedent raw text; detailed description at
+  -- 0.14.2: https://github.com/typst/typst/blob/b33de9de113c91c184214b299bd7a8eb3070c3ab/crates/typst-syntax/src/lexer.rs#L310-L338 
+  
+  -- Use parser to handle CRLF; built-in `lines` would also strip trailing \n
+  let (most, lst) = either (\_ -> ([], "")) id $ runParser splitlines () "" code
+  let nonblankSkipFirst = filter (not . all isSpace) $ drop 1 most
+  -- The last line is always considered
+  let dedent = minOrZero $ map (length . takeWhile isSpace) $ lst:nonblankSkipFirst
+  
+  let mostDedented = case most of
+                        [] -> []
+                        (x:xs) | all isSpace x -> map (drop dedent) xs
+                        -- remove optional single space after the language tag/opening ```
+                        ((' ':x):xs) -> x:(map (drop dedent) xs) 
+                        (x:xs) -> x:(map (drop dedent) xs) 
+  let dedented = if not (all isSpace lst)
+                then mostDedented <> [stripOneSpaceIfEndsWithBacktick $ drop dedent lst]
+                else mostDedented
+
+  pure $ RawBlock lang (T.pack $ intercalate "\n" dedented)
   where
-    gobble :: Int -> P ()
-    gobble 0 = pure ()
-    gobble n = (char ' ' *> gobble (n - 1)) <|> pure ()
+    -- This parser should always succeed
+    splitlines :: Parsec String () ([String], String)
+    splitlines = do
+      most <- many (try $ manyTill anyChar endOfLine)
+      lst <- manyTill anyChar eof
+      pure (most, lst)
+    minOrZero [] = 0
+    minOrZero xs = minimum xs
+    -- Allowed lang tag might change in future Typst versions, see
+    -- https://github.com/typst/typst/pull/7337
+    pLangTag = do
+      c <- satisfy isIdentStart
+      cs <- many $ satisfy isIdentContinue
+      pure $ (c : cs)
+    stripOneSpaceIfEndsWithBacktick s = case dropWhileEnd isSpace s of
+        [] -> s
+        xs | last xs == '`' -> dropOneSpaceAtEnd s
+        _ -> s
+      where
+        dropOneSpaceAtEnd xs =
+          case reverse xs of
+            (' ' : rest) -> reverse rest
+            _            -> xs
 
 pStrong :: P Markup
-pStrong = Strong <$> (char '*' *> manyTill pMarkup (char '*'))
+pStrong = do
+  void $ char '*'
+  updateState $ \st -> st {stAtStart = False}
+  Strong <$> manyTill pMarkup (char '*')
 
 pEmph :: P Markup
-pEmph = Emph <$> (char '_' *> manyTill pMarkup (char '_'))
+pEmph = do
+  void $ char '_'
+  updateState $ \st -> st {stAtStart = False}
+  Emph <$> manyTill pMarkup (char '_')
 
 pHeading :: P Markup
 pHeading = try $ do
-  col <- sourceColumn <$> getPosition
-  lineStartCol <- stLineStartCol <$> getState
-  guard (col == lineStartCol)
+  getState >>= guard . stAtStart
   lev <- length <$> many1 (char '=')
-  void (many1 (char ' ')) <|> void (lookAhead endOfLine)
-  -- Note: == hi _foo
-  -- bar_ is parsed as a heading with "hi emph(foobar)"
-  ms <- manyTill pMarkup (    void pEol
-                          <|> pEndOfContent
-                          <|> void (lookAhead (try (spaces *> pLabel)))
-                          <|> void (lookAhead (char ']')))
-  skipMany spaceChar
-  pure $ Heading lev ms
+  guardAfterListOrHeadingMarker
+  updateState $ \st -> st {stAtStart = False}
+  Heading lev <$> pSinglelineWhileNot pLabel
 
 pUrl :: P Markup
 pUrl = try $ do
@@ -846,7 +891,8 @@ pBaseExpr ident_parser =
     <|> ident_parser
     <|> pArrayExpr
     <|> pDictExpr
-    <|> inParens pExpr
+    <|> try (inParens pExpr)
+    <|> (Binding <$> pDestructuringBind True)
     <|> pLabel
     <|> (Block . Content . (: [])
          <$> lexeme (pRawBlock <|> pRawInline <|> pEquation))
@@ -1040,25 +1086,25 @@ pCode = sepEndBy pExpr (void (sym ";") <|> ws)
 -- content-block ::= '[' markup ']'
 pContent :: P Block
 pContent = do
+  pos <- getPosition
   void $ char '['
-  col <- sourceColumn <$> getPosition
-  oldLineStartCol <- stLineStartCol <$> getState
-  oldIndent <- stIndent <$> getState
+  oldBracketNesting <- stBracketNesting <$> getState
   updateState $ \st ->
     st
-      { stLineStartCol = col,
+      { stAtStart = True,
         stContentBlockNesting =
           stContentBlockNesting st + 1,
-        stIndent = []
+        stBracketNesting = 0
       }
-  ms <- manyTill pMarkup (char ']')
+  ms <- many pMarkup
+  void $ (char ']' <?> "unclosed delimiter at" <> show pos)
   ws
   updateState $ \st ->
     st
-      { stLineStartCol = oldLineStartCol,
+      { stAtStart = False,
         stContentBlockNesting =
           stContentBlockNesting st - 1,
-        stIndent = oldIndent
+        stBracketNesting = oldBracketNesting
       }
   pure $ Content ms
 
@@ -1070,7 +1116,7 @@ pEndOfContent =
       then void (lookAhead (char ']'))
       else mzero
 
--- array-expr ::= '(' ((expr ',') | (expr (',' expr)+ ','?))? ')'
+-- array-expr ::= '(' (('..' expr) | (expr ',') | (expr (',' expr)+ ','?))? ')'
 pArrayExpr :: P Expr
 pArrayExpr =
   try $
@@ -1078,9 +1124,9 @@ pArrayExpr =
       ( do
           v <- pSpread <|> (Reg <$> pExpr)
           vs <- many $ try $ sym "," *> (pSpread <|> (Reg <$> pExpr))
-          if null vs
-            then void $ sym ","
-            else optional $ void $ sym ","
+          case (v, vs) of
+            ((Reg _), []) -> void $ sym ","
+            _ -> optional $ void $ sym ","
           pure $ Array (v : vs)
       )
         <|> (Array [] <$ optional (void $ sym ","))
@@ -1165,11 +1211,20 @@ pParam =
       i <- try pIdentifier
       (DefaultParam i <$> (sym ":" *> pExpr)) <|> pure (NormalParam i)
     pDestructuringParam = do
-      DestructuringBind parts <- pDestructuringBind
+      DestructuringBind parts <- pDestructuringBind False
       pure $ DestructuringParam parts
 
-pBind :: P Bind
-pBind = pBasicBind <|> pDestructuringBind
+pBind :: Bool -> P Bind
+pBind False = pBasicBind <|> pDestructuringBind False
+-- arbitrary expressions are allowed in assignment binds but DestructuringBind is preferred
+pBind True = pDestructuringBind True <|> pExprBind
+  where
+    pExprBind = do
+      expr <- pExpr
+      pure $ case expr of
+        Underscore -> BasicBind Nothing
+        Ident ident -> BasicBind $ Just ident
+        _ -> ExprBind expr
 
 pBasicBind :: P Bind
 pBasicBind = BasicBind <$> try (pBindIdentifier <|> inParens pBindIdentifier)
@@ -1181,30 +1236,34 @@ pBindIdentifier = do
      then pure Nothing
      else pure $ Just ident
 
-pDestructuringBind :: P Bind
-pDestructuringBind =
+pDestructuringBind :: Bool -> P Bind
+pDestructuringBind allowExpr =
   inParens $
-    DestructuringBind <$> (pBindPart `sepEndBy` (sym ","))
+    DestructuringBind <$> ((pSink <|> pWithKey <|> pSimple) `sepEndBy` (sym ","))
   where
-    pBindPart = do
-      sink <- option False $ True <$ string ".."
-      if sink
+    pSink = do
+      void $ string ".."
+      if allowExpr
         then do
-          ident <- option Nothing pBindIdentifier -- ..
-          pure $ Sink ident
-        else do
-          ident <- pBindIdentifier
-          case ident of
-            Nothing -> pure (Simple ident)
-            Just key ->
-              (WithKey key <$> (sym ":" *> pBindIdentifier))
-                <|> pure (Simple ident)
+          mbexpr <- option Nothing $ Just <$> pExpr
+          pure $ case mbexpr of
+            Nothing -> Sink Nothing
+            Just Underscore -> Sink Nothing
+            Just (Ident ident) -> Sink $ Just ident
+            Just (expr) -> ExprSink expr
+        else Sink <$> option Nothing pBindIdentifier
+    pWithKey = do
+      key <- try $ pBindIdentifier <* sym ":"
+      case key of
+        Nothing -> fail "expected identifier, found underscore"
+        Just ident -> WithKey ident <$> pBind allowExpr
+    pSimple = Simple <$> pBind allowExpr
 
 -- let-expr ::= 'let' ident params? '=' expr
 pLetExpr :: P Expr
 pLetExpr = do
   pKeyword "let"
-  bind <- pBind
+  bind <- pBind False
   case bind of
     BasicBind mbname -> do
       mbparams <- option Nothing $ Just <$> pParams
@@ -1255,7 +1314,7 @@ pWhileExpr = pKeyword "while" *> (While <$> pExpr <*> pBlock)
 -- for-expr ::= 'for' bind 'in' expr block
 pForExpr :: P Expr
 pForExpr =
-  pKeyword "for" *> (For <$> pBind <*> (pKeyword "in" *> pExpr) <*> pBlock)
+  pKeyword "for" *> (For <$> pBind False <*> (pKeyword "in" *> pExpr) <*> pBlock)
 
 pImportExpr :: P Expr
 pImportExpr = pKeyword "import" *> (Import <$> pExpr <*> pImportItems)
